/*
 * UVC gadget test application
 *
 * Copyright (C) 2010 Ideas on board SPRL <laurent.pinchart@ideasonboard.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 */

#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include <linux/usb/ch9.h>
#include <linux/usb/video.h>
#include <linux/videodev2.h>

#include <sc_lib.h>
#include <sc_video.h>

#include <libguvc.h>

struct uvc_xarina {
	uint32_t tid;
	SC_VENC_INI_YUV_INFO yuv_info;
	SC_VENC_INI_JPEG_INFO jpeg_info;
	SC_VENC_INI_H264_INFO h264_info;
};

static SC_VENC_INI_YUV yuv_def_param = {
	.threadId = 0,
	.frameRate = {
		.frameRateNum = 15,
		.frameRateDenom = 1,
	},
	.bufferFrameNum = 2,
	.releaseMode = 0,
	.inputSource = 0,
	.outImage = {
		.imageParam = {
			.width = 640,
			.height = 360,
			.frameWidth = 640,
			.frameHeight = 360,
			.stride = 640 * 2,
			.sliceHeight = 360,
			.imageFormat = 0x1b,/* UYVY */
			.videoRange = 0,
		},
		.rotation = 0,
		.invert = 0,
	},
};

static SC_VENC_INI_JPEG jpeg_def_param = {
	.threadId = 0,
	.frameRate = {
		.frameRateNum = 15,
		.frameRateDenom = 1,
	},
	.bufferFrameNum = 2,
	.releaseMode = 0,
	.inputSource = 0,
	.outImage = {
		.imageParam = {
			.width = 640,
			.height = 360,
			.frameWidth = 640,
			.frameHeight = 360,
			.stride = 640 * 2,
			.sliceHeight = 360,
			.imageFormat = 0x1b,/* UYVY */
			.videoRange = 0,
		},
		.rotation = 0,
		.invert = 0,
	},
	.codecParam = {
		.qScale = 30,
		.driNum = 0,		
	},
	.trackingRate = 0,
	.trackingUnderLimit = 1,
};

static SC_VENC_INI_H264 h264_def_param = {
	.threadId = 0,
	.frameRate = {
		.frameRateNum = 15,
		.frameRateDenom = 1,
	},
	.bufferFrameNum = 2,
	.releaseMode = 0,
	.inputSource = 0,
	.outImage = {
		.imageParam = {
			.width = 640,
			.height = 360,
			.frameWidth = 640,
			.frameHeight = 360,
			.stride = 640 * 2,
			.sliceHeight = 360,
			.imageFormat = 0x7f00008A,
			.videoRange = 0,
		},
		.rotation = 0,
		.invert = 0,
	},
	.codecParam = {
		.targetBitrate = 8000000,
		.maxBitrate = 16000000,
		.profile = 0,
		.level = 13,
		.entropyCodingMode = 0,
		.nNd = 30,
		.nN = 30,
		.nM = 1,
		.insertHeaderMode = 2,
		.deblockingAlpha = -6,
		.deblockingBeta = -6,
		.rateControlParam = {
			.maxQpPicI = 51,
			.maxQpPicP = 51,
			.maxQpPicB = 51,
			.minQpPicI = 10,
			.minQpPicP = 10,
			.minQpPicB = 10,
		},
	},
};

static void usage(const char *argv0)
{
	fprintf(stderr, "Usage: %s [options]\n", argv0);
	fprintf(stderr, "Available options are\n");
	fprintf(stderr, " -b		Use bulk mode\n");
	fprintf(stderr, " -d device	Video device\n");
	fprintf(stderr, " -h		Print this help screen and exit\n");
	fprintf(stderr, " -i image	MJPEG image\n");
}

static int uvc_xarina_fill_buffer(struct uvc_device *dev, void *buf)
{
	struct uvc_xarina *xa_uvc;
	SC_VENC_ENC_DATA data;

	xa_uvc = uvc_device_to_priv(dev);
	if (!xa_uvc)
		return -1;

	if (sc_venc_get_stream_data(xa_uvc->tid, -1, &data) != 0)
		return 0;

	memcpy(buf, data.shm.pStreamAdr, data.shm.streamSize);
	sc_venc_release_stream_data(xa_uvc->tid, &data.shm);

	return data.shm.streamSize;
}

int uvc_xarina_start_stream(struct uvc_device *dev,
							__u32 fcc, __u32 width,
							__u32 height, __u32 interval)
{
	struct uvc_xarina *xa_uvc;

	xa_uvc = uvc_device_to_priv(dev);
	if (!xa_uvc)
		return -1;

	switch (fcc) {
	case V4L2_PIX_FMT_YUYV:
		yuv_def_param.outImage.imageParam.width = width;
		yuv_def_param.outImage.imageParam.frameWidth = width;
		yuv_def_param.outImage.imageParam.stride = (width * 2);
		yuv_def_param.outImage.imageParam.height = height;
		yuv_def_param.outImage.imageParam.frameHeight = height;
		yuv_def_param.outImage.imageParam.sliceHeight = height;
		yuv_def_param.frameRate.frameRateNum = 10000000 / interval;
		yuv_def_param.frameRate.frameRateDenom = 1;
		if (sc_venc_open_yuv_session(&yuv_def_param,
									 &xa_uvc->tid,
									 &xa_uvc->yuv_info) < 0) {
			printf("sc_venc_open_yuv_session failed.[errno=0x%08x]\n", errno);
			printf("width=%d width=%d %dfps\n", width, height, (10000000/interval));
			return -1;
		}
		printf("call sc_venc_open_yuv_session(%dx%d@%dfps)\n", width, height, (10000000/interval));
		break;
	case V4L2_PIX_FMT_MJPEG:
		jpeg_def_param.outImage.imageParam.width = width;
		jpeg_def_param.outImage.imageParam.frameWidth = width;
		jpeg_def_param.outImage.imageParam.stride = (width * 2);
		jpeg_def_param.outImage.imageParam.height = height;
		
		if(height==1080){
			jpeg_def_param.outImage.imageParam.frameHeight = 1088;
			jpeg_def_param.outImage.imageParam.sliceHeight = 1088;
		}else{
			jpeg_def_param.outImage.imageParam.frameHeight = height;
			jpeg_def_param.outImage.imageParam.sliceHeight = height;
		}
		
		jpeg_def_param.frameRate.frameRateNum = 10000000 / interval;
		jpeg_def_param.frameRate.frameRateDenom = 1;
		if (sc_venc_open_jpeg_session(&jpeg_def_param,
									 &xa_uvc->tid,
									 &xa_uvc->jpeg_info) < 0) {
			printf("sc_venc_open_jpeg_session failed.[errno=0x%08x]\n", errno);
			printf("width=%d width=%d %dfps\n", width, height, (10000000/interval));
			return -1;
		}
		printf("call sc_venc_open_jpeg_session(%dx%d@%dfps)\n", width, height, (10000000/interval));
		break;
	case V4L2_PIX_FMT_H264:
		h264_def_param.outImage.imageParam.width = width;
		h264_def_param.outImage.imageParam.frameWidth = width;
		h264_def_param.outImage.imageParam.stride = (width * 2);
		h264_def_param.outImage.imageParam.height = height;
		h264_def_param.outImage.imageParam.frameHeight = height;
		h264_def_param.outImage.imageParam.sliceHeight = height;
		h264_def_param.frameRate.frameRateNum = 10000000 / interval;
		h264_def_param.frameRate.frameRateDenom = 1;
		if (sc_venc_open_h264_session(&h264_def_param,
									 &xa_uvc->tid,
									 &xa_uvc->h264_info) < 0) {
			printf("sc_venc_open_h264_session failed.[errno=0x%08x]\n", errno);
			printf("width=%d width=%d %dfps\n", width, height, (10000000/interval));
			return -1;
		}
		printf("call sc_venc_open_h264_session(%dx%d@%fps)\n", width, height, (10000000/interval));
		break;
	default:
		printf("not support fcc(%x)\n", fcc);
		return -1;
	}

	if (sc_venc_start(xa_uvc->tid) < 0) {
		printf("sc_venc_start failed.\n");
		sc_venc_close(xa_uvc->tid);
		return -1;
	}

	return 0;
}

int uvc_xarina_stop_stream(struct uvc_device *dev)
{
	struct uvc_xarina *xa_uvc;
	SC_SHM_STR_POS shm;

	xa_uvc = uvc_device_to_priv(dev);
	if (!xa_uvc)
		return -1;

	if (sc_venc_stop(xa_uvc->tid, &shm) < 0)
		printf("sc_venc_stop failed.\n");

	if (sc_venc_close(xa_uvc->tid) < 0)
		printf("sc_venc_close failed.\n");

	return 0;
}

static struct uvc_ops uvc_xarina_ops = {
	.fill_buffer = uvc_xarina_fill_buffer,
	.start_stream = uvc_xarina_start_stream,
	.stop_stream = uvc_xarina_stop_stream,
};

int main(int argc, char *argv[])
{
	char *device = "/dev/video0";
	struct uvc_device *dev;
	int opt;
	struct uvc_xarina *xa_uvc;

	while ((opt = getopt(argc, argv, "bd:h")) != -1) {
		switch (opt) {
		case 'd':
			device = optarg;
			break;

		case 'h':
			usage(argv[0]);
			return 0;

		default:
			fprintf(stderr, "Invalid option '-%c'\n", opt);
			usage(argv[0]);
			return 1;
		}
	}

	xa_uvc = malloc(sizeof(struct uvc_xarina));
	if (!xa_uvc) {
		printf("struct uvc_xarina malloc failed.\n");
		return 1;
	}

	dev = uvc_open(device);
	if (dev == NULL) {
		printf("uvc_open failed.\n");
		goto Err;
	}

	uvc_register_private_data(dev, xa_uvc);

	if (sc_video_init() < 0) {
		printf("sc_video_init failed.\n");
		goto Err;
	}

	uvc_register_ops(dev, &uvc_xarina_ops);

	uvc_start(dev);
	uvc_close(dev);

	sc_video_exit();
	free(xa_uvc);

	return 0;
 Err:
	sc_video_exit();
	free(xa_uvc);
	return 1;
}

